From 6ad3c9750560f9776a553074cb39cf4506cf0634 Mon Sep 17 00:00:00 2001
From: Christophe Courtaut <christophe.courtaut@gmail.com>
Date: Fri, 31 Jul 2009 15:59:09 +0200
Subject: [PATCH] Functionnal version of smem stream output module

This patch contains a smem stream output module, which allows you
to render using memory buffer, with pace_out_control or not.
Code was inspired by mosaic_bridge mainly, and for now, only video
output is available.
---
 modules/stream_out/Modules.am |    2 +
 modules/stream_out/smem.c     |  495 +++++++++++++++++++++++++++++++++++++++++
 2 files changed, 497 insertions(+), 0 deletions(-)
 create mode 100644 modules/stream_out/smem.c

diff --git a/modules/stream_out/Modules.am b/modules/stream_out/Modules.am
index fd09f9e..36625b3 100644
--- a/modules/stream_out/Modules.am
+++ b/modules/stream_out/Modules.am
@@ -12,6 +12,7 @@ SOURCES_stream_out_mosaic_bridge = mosaic_bridge.c
 SOURCES_stream_out_autodel = autodel.c
 SOURCES_stream_out_record = record.c
 SOURCES_stream_out_raop = raop.c
+SOURCES_stream_out_smem = smem.c
 
 libvlc_LTLIBRARIES += \
 	libstream_out_dummy_plugin.la \
@@ -26,6 +27,7 @@ libvlc_LTLIBRARIES += \
 	libstream_out_mosaic_bridge_plugin.la \
 	libstream_out_autodel_plugin.la \
 	libstream_out_record_plugin.la \
+	libstream_out_smem_plugin.la \
 	$(NULL)
 
 if HAVE_LIBGCRYPT
diff --git a/modules/stream_out/smem.c b/modules/stream_out/smem.c
new file mode 100644
index 0000000..a8dbc65
--- /dev/null
+++ b/modules/stream_out/smem.c
@@ -0,0 +1,495 @@
+/*****************************************************************************
+ * smem.c: stream output to memory buffer module
+ *****************************************************************************
+ * Copyright (C) 2009 the VideoLAN team
+ * $Id$
+ *
+ * Authors: Christophe Courtaut <christophe.courtaut@gmail.com>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
+ *****************************************************************************/
+
+/*****************************************************************************
+ * Preamble
+ *****************************************************************************/
+
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include <vlc_common.h>
+#include <vlc_plugin.h>
+#include <vlc_input.h>
+#include <vlc_sout.h>
+#include <vlc_block.h>
+#include <vlc_codec.h>
+#include <vlc_picture.h>
+#include <vlc_image.h>
+
+/*****************************************************************************
+ * Module descriptor
+ *****************************************************************************/
+#define T_VIDEO_WIDTH N_( "Video Width" )
+#define LT_VIDEO_WIDTH N_( "Video memory buffer width." )
+
+#define T_VIDEO_HEIGHT N_( "Video Height" )
+#define LT_VIDEO_HEIGHT N_( "Video memory buffer height." )
+
+#define T_VIDEO_PITCH N_( "Video Pitch" )
+#define LT_VIDEO_PITCH N_( "Video memory buffer pitch in bytes." )
+
+#define T_VIDEO_CHROMA N_( "Video Chroma" )
+#define LT_VIDEO_CHROMA N_( "Output chroma for the memory image as a 4-character " \
+                      "string, eg. \"RV32\"." )
+
+#define T_VIDEO_PRERENDER_CALLBACK N_( "Video prerender callback" )
+#define LT_VIDEO_PRERENDER_CALLBACK N_( "Address of the video prerender callback function" \
+                                "this function will set the buffer where render will be done" )
+
+#define T_AUDIO_PRERENDER_CALLBACK N_( "Audio prerender callback" )
+#define LT_AUDIO_PRERENDER_CALLBACK N_( "Address of the audio prerender callback function." \
+                                        "this function will set the buffer where render will be done" )
+
+#define T_VIDEO_POSTRENDER_CALLBACK N_( "Video postrender callback" )
+#define LT_VIDEO_POSTRENDER_CALLBACK N_( "Address of the video postrender callback function." \
+                                        "this function will be called when the render is into the buffer" )
+
+#define T_AUDIO_POSTRENDER_CALLBACK N_( "Audio postrender callback" )
+#define LT_AUDIO_POSTRENDER_CALLBACK N_( "Address of the audio postrender callback function." \
+                                        "this function will be called when the render is into the buffer" )
+
+#define T_VIDEO_DATA N_( "Video Callback data" )
+#define LT_VIDEO_DATA N_( "Data for the video callback function." )
+
+#define T_AUDIO_DATA N_( "Audio callback data" )
+#define LT_AUDIO_DATA N_( "Data for the audio callback function." )
+
+#define T_TIME_SYNC N_( "Time Synchronized output" )
+#define LT_TIME_SYNC N_( "Time Synchronisation option for output. " \
+                        "If true, stream will render as usual, else " \
+                        "it will be rendered as fast as possible.")
+
+static int  Open ( vlc_object_t * );
+static void Close( vlc_object_t * );
+
+#define SOUT_CFG_PREFIX "sout-smem-"
+#define SOUT_PREFIX_VIDEO SOUT_CFG_PREFIX"video-"
+#define SOUT_PREFIX_AUDIO SOUT_CFG_PREFIX"audio-"
+
+vlc_module_begin ()
+    set_shortname( N_("smem"))
+    set_description( N_("Stream output to memory buffer") )
+    set_capability( "sout stream", 50 )
+    add_shortcut( "smem" )
+    set_category( CAT_SOUT )
+    set_subcategory( SUBCAT_SOUT_STREAM )
+    add_integer( SOUT_PREFIX_VIDEO "width", 320, NULL, T_VIDEO_WIDTH, LT_VIDEO_WIDTH, false )
+    add_integer( SOUT_PREFIX_VIDEO "height", 200, NULL, T_VIDEO_HEIGHT, LT_VIDEO_HEIGHT, false )
+    add_integer( SOUT_PREFIX_VIDEO "pitch", 640, NULL, T_VIDEO_PITCH, LT_VIDEO_PITCH, false )
+    add_string( SOUT_PREFIX_VIDEO "chroma", "RV16", NULL, T_VIDEO_CHROMA, LT_VIDEO_CHROMA, true )
+    add_string( SOUT_PREFIX_VIDEO "prerender-callback", "0", NULL, T_VIDEO_PRERENDER_CALLBACK, LT_VIDEO_PRERENDER_CALLBACK, true )
+    add_string( SOUT_PREFIX_AUDIO "prerender-callback", "0", NULL, T_AUDIO_PRERENDER_CALLBACK, LT_AUDIO_PRERENDER_CALLBACK, true )
+    add_string( SOUT_PREFIX_VIDEO "postrender-callback", "0", NULL, T_VIDEO_POSTRENDER_CALLBACK, LT_VIDEO_POSTRENDER_CALLBACK, true )
+    add_string( SOUT_PREFIX_AUDIO "postrender-callback", "0", NULL, T_AUDIO_POSTRENDER_CALLBACK, LT_AUDIO_POSTRENDER_CALLBACK, true )
+    add_string( SOUT_PREFIX_VIDEO "data", "0", NULL, T_VIDEO_DATA, LT_VIDEO_DATA, true )
+    add_string( SOUT_PREFIX_AUDIO "data", "0", NULL, T_AUDIO_DATA, LT_VIDEO_DATA, true )
+    add_bool( SOUT_CFG_PREFIX "time-sync", true, NULL, T_TIME_SYNC, LT_TIME_SYNC, true );
+    set_callbacks( Open, Close )
+vlc_module_end ()
+
+
+/*****************************************************************************
+ * Exported prototypes
+ *****************************************************************************/
+static const char *const ppsz_sout_options[] = {
+    "video-width", "video-height", "video-pitch",
+    "video-chroma", "video-prerender-callback", "audio-prerender-callback",
+    "video-postrender-callback", "audio-postrender-callback", "video-data", "audio-data", "time-sync", NULL
+};
+
+static sout_stream_id_t *Add ( sout_stream_t *, es_format_t * );
+static int               Del ( sout_stream_t *, sout_stream_id_t * );
+static int               Send( sout_stream_t *, sout_stream_id_t *, block_t* );
+
+static sout_stream_id_t *AddVideo( sout_stream_t *p_stream, es_format_t *p_fmt );
+static sout_stream_id_t *AddAudio( sout_stream_t *p_stream, es_format_t *p_fmt );
+
+inline static void video_del_buffer_decoder( decoder_t *, picture_t * );
+inline static picture_t *video_new_buffer_decoder( decoder_t * );
+static picture_t *video_new_buffer( vlc_object_t *, decoder_owner_sys_t *,
+                                    es_format_t * );
+static void video_link_picture_decoder( decoder_t *, picture_t * );
+static void video_unlink_picture_decoder( decoder_t *, picture_t * );
+
+static int SendVideo( sout_stream_t *p_stream, sout_stream_id_t *id,
+                      block_t *p_buffer );
+static int SendVideoPicture( sout_stream_t *p_stream, picture_t *p_pic );
+static int SendAudio( sout_stream_t *p_stream, sout_stream_id_t *id,
+                      block_t *p_buffer );
+
+struct sout_stream_id_t
+{
+
+};
+
+struct decoder_owner_sys_t
+{
+    /* Current format in use by the output */
+    video_format_t video;
+};
+
+struct sout_stream_sys_t
+{
+    vlc_mutex_t *p_lock;
+    decoder_t       *p_decoder;
+    image_handler_t *p_image;
+    int i_video_width, i_video_height, i_video_pitch;
+    vlc_fourcc_t i_video_chroma;
+
+    void ( *pf_video_prerender_callback ) ( void* , uint8_t** );
+    void ( *pf_audio_prerender_callback ) ( void* , void* );
+    void ( *pf_video_postrender_callback ) ( void* );
+    void ( *pf_audio_postrender_callback ) ( void*, void* );
+    void *p_audio_data;
+    void *p_video_data;
+    bool time_sync;
+};
+
+/*****************************************************************************
+ * Open:
+ *****************************************************************************/
+static int Open( vlc_object_t *p_this )
+{
+    char* psz_tmp;
+    sout_stream_t *p_stream = (sout_stream_t*)p_this;
+    sout_stream_sys_t *p_sys;
+
+    p_sys = calloc( 1, sizeof( sout_stream_sys_t ) );
+    if( !p_sys )
+        return VLC_ENOMEM;
+    p_stream->p_sys = p_sys;
+
+    /* Getting module parameters */
+    p_sys->i_video_width = var_CreateGetInteger( p_stream, SOUT_PREFIX_VIDEO "width");
+    p_sys->i_video_height = var_CreateGetInteger( p_stream, SOUT_PREFIX_VIDEO "height");
+    p_sys->i_video_pitch = var_CreateGetInteger( p_stream, SOUT_PREFIX_VIDEO "pitch");
+    p_sys->time_sync = var_CreateGetBool( p_stream, SOUT_CFG_PREFIX "time-sync" );
+
+    psz_tmp = var_CreateGetString( p_stream, SOUT_PREFIX_VIDEO "chroma" );
+    p_sys->i_video_chroma = vlc_fourcc_GetCodecFromString( VIDEO_ES, psz_tmp );
+    free( psz_tmp );
+
+    psz_tmp = var_CreateGetString( p_stream, SOUT_PREFIX_VIDEO "prerender-callback" );
+    p_sys->pf_video_prerender_callback = (void (*) (void *, uint8_t**))(intptr_t)atoll( psz_tmp );
+    free( psz_tmp );
+
+    psz_tmp = var_CreateGetString( p_stream, SOUT_PREFIX_AUDIO "prerender-callback" );
+    p_sys->pf_audio_prerender_callback = (void (*) (void *, void*))(intptr_t)atoll( psz_tmp );
+    free( psz_tmp );
+
+    psz_tmp = var_CreateGetString( p_stream, SOUT_PREFIX_VIDEO "postrender-callback" );
+    p_sys->pf_video_postrender_callback = (void (*) (void *))(intptr_t)atoll( psz_tmp );
+    free( psz_tmp );
+
+    psz_tmp = var_CreateGetString( p_stream, SOUT_PREFIX_AUDIO "postrender-callback" );
+    p_sys->pf_audio_postrender_callback = (void (*) (void *, void*))(intptr_t)atoll( psz_tmp );
+    free( psz_tmp );
+
+    psz_tmp = var_CreateGetString( p_stream, SOUT_PREFIX_VIDEO "data" );
+    p_sys->p_video_data = (void *)( intptr_t )atoll( psz_tmp );
+    free( psz_tmp );
+
+    psz_tmp = var_CreateGetString( p_stream, SOUT_PREFIX_AUDIO "data" );
+    p_sys->p_audio_data = (void *)( intptr_t )atoll( psz_tmp );
+    free( psz_tmp );
+
+    /* Setting stream out module callbacks */
+    p_stream->pf_add    = Add;
+    p_stream->pf_del    = Del;
+    p_stream->pf_send   = Send;
+
+    /* Does the module need out_pace_control? */
+    if ( p_sys->time_sync )
+        p_stream->p_sout->i_out_pace_nocontrol++;
+
+    return VLC_SUCCESS;
+}
+
+/*****************************************************************************
+ * Close:
+ *****************************************************************************/
+static void Close( vlc_object_t * p_this )
+{
+    sout_stream_t *p_stream = (sout_stream_t*)p_this;
+    if ( p_stream->p_sys->time_sync )
+        p_stream->p_sout->i_out_pace_nocontrol--;
+    free( p_stream->p_sys );
+}
+
+static sout_stream_id_t *Add( sout_stream_t *p_stream, es_format_t *p_fmt )
+{
+    sout_stream_id_t *id = NULL;
+
+    if ( p_fmt->i_cat == VIDEO_ES )
+        id = AddVideo( p_stream, p_fmt );
+    else if ( p_fmt->i_cat == AUDIO_ES )
+        id = AddAudio( p_stream, p_fmt );
+    return id;
+}
+
+static sout_stream_id_t *AddVideo( sout_stream_t *p_stream, es_format_t *p_fmt )
+{
+    sout_stream_sys_t *p_sys = p_stream->p_sys;
+
+    /* Creation of decoder object */
+
+    p_sys->p_decoder = vlc_object_create( p_stream, VLC_OBJECT_DECODER );
+
+    if ( !p_sys->p_decoder )
+        return NULL;
+
+    vlc_object_attach( p_sys->p_decoder, p_stream );
+    p_sys->p_decoder->p_module = NULL;
+    p_sys->p_decoder->fmt_in = *p_fmt;
+    p_sys->p_decoder->b_pace_control = false;
+    p_sys->p_decoder->fmt_out = p_sys->p_decoder->fmt_in;
+    p_sys->p_decoder->fmt_out.i_extra = 0;
+    p_sys->p_decoder->fmt_out.p_extra = NULL;
+    p_sys->p_decoder->pf_decode_video = 0;
+    p_sys->p_decoder->pf_vout_buffer_new = video_new_buffer_decoder;
+    p_sys->p_decoder->pf_vout_buffer_del = video_del_buffer_decoder;
+    p_sys->p_decoder->pf_picture_link    = video_link_picture_decoder;
+    p_sys->p_decoder->pf_picture_unlink  = video_unlink_picture_decoder;
+    p_sys->p_decoder->p_owner = malloc( sizeof(decoder_owner_sys_t) );
+    if( !p_sys->p_decoder->p_owner )
+    {
+        vlc_object_detach( p_sys->p_decoder );
+        vlc_object_release( p_sys->p_decoder );
+        return NULL;
+    }
+
+    p_sys->p_decoder->p_owner->video = p_fmt->video;
+    p_sys->p_decoder->p_module =
+        module_need( p_sys->p_decoder, "decoder", "$codec", false );
+
+    if( !p_sys->p_decoder->p_module || !p_sys->p_decoder->pf_decode_video )
+    {
+        if( p_sys->p_decoder->p_module )
+        {
+            msg_Err( p_stream, "instanciated a non video decoder" );
+            module_unneed( p_sys->p_decoder, p_sys->p_decoder->p_module );
+        }
+        else
+        {
+            msg_Err( p_stream, "cannot find decoder" );
+        }
+        free( p_sys->p_decoder->p_owner );
+        vlc_object_detach( p_sys->p_decoder );
+        vlc_object_release( p_sys->p_decoder );
+        return NULL;
+    }
+    if ( p_sys->i_video_height && p_sys->i_video_width )
+        p_sys->p_image = image_HandlerCreate( p_stream );
+    else
+        p_sys->p_image = NULL;
+    return ( sout_stream_id_t* )p_sys;
+}
+
+static sout_stream_id_t *AddAudio( sout_stream_t *p_stream, es_format_t *p_fmt )
+{
+    VLC_UNUSED( p_stream );
+    VLC_UNUSED( p_fmt );
+    return NULL;
+}
+
+
+
+static int Del( sout_stream_t *p_stream, sout_stream_id_t *id )
+{
+    VLC_UNUSED(p_stream);
+    free( id );
+    return VLC_SUCCESS;
+}
+
+static int Send( sout_stream_t *p_stream, sout_stream_id_t *id,
+                 block_t *p_buffer )
+{
+    sout_stream_sys_t *p_sys = p_stream->p_sys;
+
+    if ( (sout_stream_sys_t *)id != p_sys )
+    {
+        block_ChainRelease( p_buffer );
+        return VLC_SUCCESS;
+    }
+
+    if ( p_sys->p_decoder->fmt_in.i_cat == VIDEO_ES )
+        return SendVideo( p_stream, id, p_buffer );
+    else if ( p_sys->p_decoder->fmt_in.i_cat == AUDIO_ES )
+        return SendAudio( p_stream, id, p_buffer );
+    return VLC_SUCCESS;
+}
+
+static int SendVideo( sout_stream_t *p_stream, sout_stream_id_t *id,
+                      block_t *p_buffer )
+{
+    VLC_UNUSED( id );
+    sout_stream_sys_t *p_sys = p_stream->p_sys;
+    picture_t *p_pic;
+
+    while ( ( p_pic = p_sys->p_decoder->pf_decode_video( p_sys->p_decoder,
+                                                         &p_buffer ) ) )
+    {
+        picture_t *p_new_pic;
+
+        /* if there is a specified width, height or chroma, we need to convert the image */
+        if( p_sys->i_video_height || p_sys->i_video_width  ||
+            ( p_sys->i_video_chroma &&
+              p_sys->i_video_chroma != p_sys->p_decoder->fmt_out.video.i_chroma ) )
+        {
+            video_format_t fmt_out, fmt_in;
+
+            memset( &fmt_in, 0, sizeof(video_format_t) );
+            memset( &fmt_out, 0, sizeof(video_format_t) );
+            fmt_in = p_sys->p_decoder->fmt_out.video;
+
+            if( p_sys->i_video_chroma )
+                fmt_out.i_chroma = p_sys->i_video_chroma;
+            else
+                fmt_out.i_chroma = VLC_CODEC_I420;
+
+            if ( p_sys->i_video_width )
+                fmt_out.i_width = p_sys->i_video_width;
+            else
+                fmt_out.i_width = fmt_in.i_width;
+            if ( p_sys->i_video_height )
+                fmt_out.i_height = p_sys->i_video_height;
+            else
+                fmt_out.i_height = fmt_in.i_visible_height;
+
+            fmt_out.i_visible_width = fmt_out.i_width;
+            fmt_out.i_visible_height = fmt_out.i_height;
+
+            p_new_pic = image_Convert( p_sys->p_image,
+                                       p_pic, &fmt_in, &fmt_out );
+            if( p_new_pic == NULL )
+            {
+                msg_Err( p_stream, "image conversion failed" );
+                picture_Release( p_pic );
+                continue;
+            }
+            picture_Release( p_pic );
+        }
+        else
+            p_new_pic = p_pic;
+        SendVideoPicture( p_stream, p_new_pic );
+        picture_Release( p_new_pic );
+    }
+    return VLC_SUCCESS;
+}
+
+static int SendVideoPicture( sout_stream_t *p_stream, picture_t *p_pic )
+{
+    uint8_t* pixels;
+
+    /* Calling prerender callback to get pixel buffer where to render */
+    p_stream->p_sys->pf_video_prerender_callback( p_stream->p_sys->p_video_data, &pixels );
+    /* Copying pixels data from picture to pixel buffer */
+    for ( int line = 0; line < p_pic->p->i_visible_lines; line++, pixels += p_pic->p->i_visible_pitch )
+        vlc_memcpy( pixels, p_pic->p->p_pixels + p_pic->p->i_pitch * line , p_pic->p->i_visible_pitch );
+    /* Calling postrender callback to tell thrid party programs that pixel buffer is now available */
+    p_stream->p_sys->pf_video_postrender_callback( p_stream->p_sys->p_video_data );
+    return VLC_SUCCESS;
+}
+
+static int SendAudio( sout_stream_t *p_stream, sout_stream_id_t *id,
+                      block_t *p_buffer )
+{
+    VLC_UNUSED( id );
+    VLC_UNUSED( p_buffer );
+    VLC_UNUSED( p_stream );
+    return VLC_SUCCESS;
+}
+
+/**********************************************************************
+ * Video Buffer Helpers
+ **********************************************************************/
+
+static inline picture_t *video_new_buffer_decoder( decoder_t *p_dec )
+{
+    return video_new_buffer( VLC_OBJECT( p_dec ),
+                             (decoder_owner_sys_t *)p_dec->p_owner,
+                             &p_dec->fmt_out );
+}
+
+static picture_t *video_new_buffer( vlc_object_t *p_this,
+                                    decoder_owner_sys_t *p_sys,
+                                    es_format_t *fmt_out )
+{
+    VLC_UNUSED( p_this );
+    if( fmt_out->video.i_width != p_sys->video.i_width ||
+        fmt_out->video.i_height != p_sys->video.i_height ||
+        fmt_out->video.i_chroma != p_sys->video.i_chroma ||
+        fmt_out->video.i_aspect != p_sys->video.i_aspect )
+    {
+        if( !fmt_out->video.i_sar_num ||
+            !fmt_out->video.i_sar_den )
+        {
+            fmt_out->video.i_sar_num =
+                fmt_out->video.i_aspect * fmt_out->video.i_height;
+
+            fmt_out->video.i_sar_den =
+                VOUT_ASPECT_FACTOR * fmt_out->video.i_width;
+        }
+
+        vlc_ureduce( &fmt_out->video.i_sar_num,
+                     &fmt_out->video.i_sar_den,
+                     fmt_out->video.i_sar_num,
+                     fmt_out->video.i_sar_den, 0 );
+
+        if( !fmt_out->video.i_visible_width ||
+            !fmt_out->video.i_visible_height )
+        {
+            fmt_out->video.i_visible_width = fmt_out->video.i_width;
+            fmt_out->video.i_visible_height = fmt_out->video.i_height;
+        }
+
+        fmt_out->video.i_chroma = fmt_out->i_codec;
+        p_sys->video = fmt_out->video;
+    }
+    fmt_out->video.i_chroma = fmt_out->i_codec;
+    return picture_New( fmt_out->video.i_chroma,
+                        fmt_out->video.i_width,
+                        fmt_out->video.i_height,
+                        fmt_out->video.i_aspect );
+}
+
+static inline void video_del_buffer_decoder( decoder_t *p_this,
+                                             picture_t *p_pic )
+{
+    VLC_UNUSED(p_this);
+    picture_Release( p_pic );
+}
+
+static void video_link_picture_decoder( decoder_t *p_dec, picture_t *p_pic )
+{
+    VLC_UNUSED(p_dec);
+    picture_Hold( p_pic );
+}
+
+static void video_unlink_picture_decoder( decoder_t *p_dec, picture_t *p_pic )
+{
+    VLC_UNUSED(p_dec);
+    picture_Release( p_pic );
+}
-- 
1.6.3.3

